/* 
 * DinX access library routines
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):	Ben Williamson <benw@pobox.com>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <poll.h>

#include <dinx/access.h>
#include <dinx/messages.h>
#include <linux/dinx.h>

#ifdef NDEBUG
#define RETURN(x) do { return (x); } while (0)
#else
#define RETURN(x) do {				\
	int __result = (x);			\
	if (__result < 0) {			\
		perror(__PRETTY_FUNCTION__);	\
	}					\
	return __result;			\
} while (0)
#endif

static const char *default_device = "/dev/dinxwin";
static const char *device_envvar = "DINXWIN";

int dinx_open(const char *device)
{
	if (!device) {
		device = getenv(device_envvar);
	}
	if (!device) {
		device = default_device;
	}
	RETURN(open(device, O_RDWR));
}

int dinx_close(int fd)
{
	RETURN(close(fd));
}

int dinx_read_event(int fd, struct devent_ops *ops, void *data)
{
	union devent event;
	union dmessage *message;
	int err = 0;
	
	err = read(fd, &event, sizeof(union devent));
	if (err < 0) {
		RETURN(err);
	}
	switch (event.type) {
	case DEVENT_REDRAW:
		if (ops->redraw) {
			ops->redraw(data, &event.redraw.rect,
					event.redraw.more);
		}
		break;
	case DEVENT_MESSAGE:
		message = (union dmessage *)event.message.data;
		switch (message->type) {
		case DMESSAGE_MOUSE:
			if (ops->mouse) {
				ops->mouse(data, &message->mouse.pos,
						message->mouse.buttons);
			}
			break;
		case DMESSAGE_KEY:
			if (ops->key) {
				ops->key(data, message->key.key);
			}
			break;
		case DMESSAGE_TREE_INFO:
			if (ops->tree_info) {
				ops->tree_info(data, &message->tree_info.bounds,
						message->tree_info.bpp);
			}
			break;
		default:
			if (ops->message) {
				ops->message(data, message);
			}
			break;
		}
		break;
	case DEVENT_SVR_CLOSED:
		printf("server closed!\n");
		break;
	default:
		err = event.type;
		break;
	}
	return err;
}

int dinx_message(int fd, const union dmessage *message)
{
	union dcommand command;
	
	command.type = DCOMMAND_MESSAGE;
	memcpy(command.message.data, message, DINX_MESSAGE_SIZE);
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_simple_message(int fd, int type)
{
	union dcommand command;
	union dmessage *message;
	
	command.type = DCOMMAND_MESSAGE;
	message = (union dmessage *)command.message.data;
	message->type = type;
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_fill_rect(int fd, const drect *rect, unsigned long col)
{
	union dcommand command;
	
	command.type = DCOMMAND_FILL_RECT;
	command.fill_rect.rect = *rect;
	command.fill_rect.col = col;
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_move_rect(int fd, const drect *rect, const dvect *moveby)
{
	union dcommand command;
	
	command.type = DCOMMAND_MOVE_RECT;
	command.move_rect.rect = *rect;
	command.move_rect.moveby = *moveby;
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_copy_rect(int fd, const drect *dest, const dvect *base,
	const void *src, int src_width, int src_bpp, const unsigned long *pal)
{
	union dcommand command;
	
	command.type = DCOMMAND_COPY_RECT;
	command.copy_rect.dest = *dest;
	command.copy_rect.base = *base;
	command.copy_rect.src = src;
	command.copy_rect.src_width = src_width;
	command.copy_rect.src_bpp = src_bpp;
	command.copy_rect.pal = pal;
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_req_bounds(int fd, const drect *bounds, const dvect *origin)
{
	union dcommand command;
	union dmessage *message;
	
	command.type = DCOMMAND_MESSAGE;
	message = (union dmessage *)command.message.data;
	message->type = DMESSAGE_REQ_BOUNDS;
	message->req_bounds.bounds = *bounds;
	message->req_bounds.origin = *origin;
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_req_map(int fd)
{
	return dinx_simple_message(fd, DMESSAGE_REQ_MAP);
}

int dinx_req_unmap(int fd)
{
	return dinx_simple_message(fd, DMESSAGE_REQ_UNMAP);
}

int dinx_req_raise(int fd)
{
	return dinx_simple_message(fd, DMESSAGE_REQ_RAISE);
}

int dinx_req_lower(int fd)
{
	return dinx_simple_message(fd, DMESSAGE_REQ_LOWER);
}

int dinx_message_to(int fd, const union dmessage *message,
					unsigned long win)
{
	union dcommand command;
	
	command.type = DCOMMAND_MESSAGE;
	command.message.win = win;
	memcpy(command.message.data, message, DINX_MESSAGE_SIZE);
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_simple_command(int fd, int type)
{
	union dcommand command;
	
	command.type = type;
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

int dinx_select_fb(int fd, int framebuffer)
{
	union dcommand command;
	
	command.type = DCOMMAND_SELECT_FB;
	command.select_fb.framebuffer = framebuffer;
	RETURN(write(fd, &command, sizeof(union dcommand)));
}

