/* 
 * DinX blitter module for 32bpp linear packed framebuffers
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):     Ben Williamson <benw@pobox.com>
 *                     Matthew Backes <lucca@acm.org>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#include <linux/module.h>
#include <linux/string.h>
#include <linux/tty.h>
#include <linux/fb.h>
#include <linux/dinx.h>
#include <linux/malloc.h>
#include <asm/uaccess.h>
#include <video/fbcon.h>	/* for mymemmove */

static int
l32_can_handle(struct display *disp)
{
	return (disp->var.bits_per_pixel == 32) &&
		(disp->type == FB_TYPE_PACKED_PIXELS);
}

static void
l32_open(void)
{
	MOD_INC_USE_COUNT;
}

static void
l32_close(void)
{
	MOD_DEC_USE_COUNT;
}

static void
l32_fill_rect(struct display *disp, const drect *rect, unsigned long col)
{
	char *line;
	int width, line_length, rows;
	drect r = *rect;

	line_length = disp->line_length;
	line = disp->screen_base + r.y1*line_length + r.x1*4;
	width = r.x2 - r.x1;
	rows = r.y2 - r.y1;
	while (rows--) {
		int cols = width;
		int *p = (int *)line;
		while (cols--) {
			*p++ = col;
		}
		line += line_length;
	}
}

static void
l32_move_rect(struct display *disp, const drect *rect, dvect moveby)
{
	drect r = *rect;
	int width, height, line_length;
	char *src, *dst;
	
	line_length = disp->line_length;
	width = (r.x2 - r.x1) * 4;
	height = r.y2 - r.y1;
	if (moveby.y < 0) {
		dst = disp->screen_base + r.y1*line_length + r.x1*4;
		src = dst - moveby.y*line_length - moveby.x*4;
		while (height--) {
			mymemmove(dst, src, width);
			dst += line_length;
			src += line_length;
		}
	} else {
		dst = disp->screen_base + r.y2*line_length + r.x1*4;
		src = dst - moveby.y*line_length - moveby.x*4;
		while (height--) {
			dst -= line_length;
			src -= line_length;
			mymemmove(dst, src, width);
		}
	}

}

static int
l32_copy_raw(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *unused)
{
	const char *s;
	char *d;
	drect r = *dest;
	int width, rows, line_length;
	
	line_length = disp->line_length;
	width = (r.x2 - r.x1) * 4;
	rows  = r.y2 - r.y1;
	s = (const char *)src + (r.x1-base->x)*4 + (r.y1-base->y)*src_width;
	d = disp->screen_base + r.y1*line_length + r.x1*4;
	while (rows--) {
		if (__copy_from_user(d, s, width)) {
			return -EFAULT;
		}
		d += line_length;
		s += src_width;
	}
	return 0;
}

static int
l32_copy_1bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	/* FIXME */
	return -ENOSYS;
}

static int
l32_copy_2bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	/* FIXME */
	return -ENOSYS;
}

static int
l32_copy_4bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	const char *s;
	char *d;
	drect r = *dest;
	int width, rows, line_length;
	
	line_length = disp->line_length;
	width = r.x2 - r.x1;
	rows  = r.y2 - r.y1;
	s = (const char *)src + r.x1-base->x + (r.y1-base->y)*src_width;
	d = disp->screen_base + r.y1*line_length + (r.x1>>1);
	while (rows--) {
		int cols = width;
		int *dp = (int *)d;
		const char *sp = s;
		while (cols) {
			unsigned char pix;
			
			if (__get_user(pix, sp)) {
			        return -EFAULT;
			}
			sp++;
			/* these might be reversed... haven't tested */
			*dp++ = palette[pix >> 4];
			*dp++ = palette[pix & 0xf];
			cols -= 2;
		}
		d += line_length;
		s += src_width;
	}
	return 0;
}

static int
l32_copy_8bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *palette)
{
	const char *s;
	char *d;
	drect r = *dest;
	int width, rows, line_length;
	
	line_length = disp->line_length;
	width = r.x2 - r.x1;
	rows  = r.y2 - r.y1;
	s = (const char *)src + r.x1-base->x + (r.y1-base->y)*src_width;
	d = disp->screen_base + r.y1*line_length + r.x1*4;
	while (rows--) {
		int cols = width;
		int *dp = (int *)d;
		const char *sp = s;
		while (cols--) {
			unsigned char pix;
			
			if (__get_user(pix, sp)) {
				return -EFAULT;
			}
			sp++;
			*dp++ = palette[pix];
		}
		d += line_length;
		s += src_width;
	}
	return 0;
}

static int
l32_copy_16bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *unused)
{
	const char *s;
	char *d;
	drect r = *dest;
	int width, rows, line_length;
	
	line_length = disp->line_length;
	width = r.x2 - r.x1;
	rows  = r.y2 - r.y1;
	s = (const char *)src + (r.x1-base->x)*2 + (r.y1-base->y)*src_width;
	d = disp->screen_base + r.y1*line_length + r.x1*4;
	while (rows--) {
		int cols = width;
		char *dp = d;
		const short *sp = (const short *)s;
		while (cols--) {
			unsigned short pix16;
			unsigned int pix32;
			
			if (__get_user(pix16, sp)) {
				return -EFAULT;
			}
			sp++;
			pix32 =  (pix16 & 0x001f) <<3;      /* 5 bits blue */
			pix32 += (pix16 & 0x07e0) >>3;      /* 6 bits green */
			pix32 += (pix16 & 0xf800) >>8;      /* 5 bits red */
			*dp++ = pix32;
		}
		d += line_length;
		s += src_width;
	}
	return 0;
}

static int
l32_copy_24bpp(struct display *disp, const drect *dest, const dvect *base,
		const void *src, int src_width, const unsigned long *unused)
{
	const char *s;
	char *d;
	drect r = *dest;
	int width = r.x2 - r.x1;
	int rows = r.y2 - r.y1;
	int line_length = disp -> line_length;
	unsigned char *buffer = kmalloc(width * 3 + 1, GFP_KERNEL);
	
	if (!buffer) {
		return -ENOMEM;
	}
	s = (const char *)src + (r.x1-base->x)*3 + (r.y1-base->y)*src_width;
	d = disp->screen_base + r.y1*line_length + r.x1*4;
        
	while (rows--) {
		int cols = width;
		int *dp = (int *)d;
	        unsigned char *sp = buffer;

		if (__copy_from_user (buffer, s, width*3)) {
			kfree(buffer);
		        return -EFAULT;
		}

		while (cols--) {
                        *dp++ = *((unsigned int*)sp) & 0xffffff; 
                        sp += 3;
		}
		d += line_length;
		s += src_width;
	}
	kfree(buffer);
	return 0;
}


static struct dinx_blitter l32_blitter = {
	NULL,		/* next */
	"linear-32bpp",
	l32_can_handle,
	l32_open,
	l32_close,
	l32_fill_rect,
	l32_move_rect,
	l32_copy_raw,
	l32_copy_1bpp,
	l32_copy_2bpp,
	l32_copy_4bpp,
	l32_copy_8bpp,
	l32_copy_16bpp,
	l32_copy_24bpp
};

int
init_dinx_l32(void)
{
	return register_blitter(&l32_blitter);
}

#ifdef MODULE

int
init_module(void)
{
	return init_dinx_l32();
}

void
cleanup_module(void)
{
	unregister_blitter(&l32_blitter);
}

#endif	/* MODULE */

