/*
 * DinX server, manages window messages and events
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is DinX.
 * 
 * The Initial Developer of the Original Code is Ben Williamson.
 * Portions created by Ben Williamson are Copyright (C) 1999
 * Ben Williamson. All Rights Reserved.
 * 
 * Contributor(s):	Ben Williamson <benw@pobox.com>
 * 
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public license (the  "[GPL] License"), in which case the
 * provisions of [GPL] License are applicable instead of those
 * above.  If you wish to allow use of your version of this file only
 * under the terms of the [GPL] License and not to allow others to use
 * your version of this file under the MPL, indicate your decision by
 * deleting  the provisions above and replace  them with the notice and
 * other provisions required by the [GPL] License.  If you do not delete
 * the provisions above, a recipient may use your version of this file
 * under either the MPL or the [GPL] License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <getopt.h>
#include <errno.h>

#include <dinx/access.h>
#include <dinx/messages.h>
#include <linux/dinx.h>
#include <linux/fb.h>

#include "vt.h"
#include "fb.h"

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

const char *usage_message =
"Usage: dinxd [--dinxsvr /dev/dinxsvr] [--framebuffer /dev/fb0]\n"
"             [--help] [--version]\n"
"\n"
"--dinxsvr <device>	Selects window tree device node\n"
"--framebuffer <device>	Selects framebuffer device node\n"
"--help			Display this message and exit\n"
"--version		Display version number and exit\n"
"\n"
"Example:  dinxd --dinxsvr /dev/dinxsvr1 --framebuffer /dev/fb2\n"
"will attach the server to DinX tree number 1, on framebuffer 2.\n"
"\n"
"On startup dinxd creates a new console and switches to it.\n"
"Unless you run some client programs after dinxd has started,\n"
"the screen will remain blank.  Chances are you can switch\n"
"back to X with LeftAlt-F7.\n"
"\n";

struct option options[] = {
	{"dinxsvr",	1, 0, 'd'},
	{"framebuffer",	1, 0, 'f'},
	{"help",	0, 0, 'h'},
	{"version",	0, 0, 'v'},
	{0,		0, 0, 0}
};

const char *svr_filename = "/dev/dinxsvr";
char *mouse_filename = "/dev/gpmdata";

struct window {
	unsigned long		win;
	struct window *		nextabove;
	struct window *		nextbelow;
	drect			bounds;
	dvect			origin;
	int			flags;
#define NEEDS_MOUSE_EVENT	(1<<0)
};

struct window *topchild;
struct window *bottomchild;

int dinx_fd, mouse_fd;
int mouse_x, mouse_y, mouse_b;

const char *framebuffer = NULL;

struct window *
find_window(unsigned long win)
{
	struct window *w;
	
	for (w = topchild; w; w = w->nextbelow) {
		if (win == w->win) {
			return w;
		}
	}
	return NULL;
}

int event_count(unsigned long win)
{
	struct dwindow_status status;
	union dcommand command;
	
	command.type = DCOMMAND_GET_STATUS;
	command.get_status.win = win;
	command.get_status.status = &status;
	write(dinx_fd, &command, sizeof(union dcommand));
	return status.event_count;
}

void send_mouse(struct window *w, int x, int y, int b)
{
	union dmessage message;
	
	message.type = DMESSAGE_MOUSE;
	message.mouse.pos.x = x;
	message.mouse.pos.y = y;
	message.mouse.buttons = b;
	dinx_message_to(dinx_fd, &message, w->win);
}

void send_tree_info(unsigned long win)
{
	union dmessage message;
	
	message.type = DMESSAGE_TREE_INFO;
	message.tree_info.bounds.x1 = 0;
	message.tree_info.bounds.y1 = 0;
	message.tree_info.bounds.x2 = fb_var.xres_virtual;
	message.tree_info.bounds.y2 = fb_var.yres_virtual;
	message.tree_info.bpp = fb_var.bits_per_pixel;
	dinx_message_to(dinx_fd, &message, win);
}

void feed_window(unsigned long win)
{
	struct window *w;
	
	w = find_window(win);
	if (!w) {
		return;
	}
	if (w->flags & NEEDS_MOUSE_EVENT) {
		w->flags &= ~NEEDS_MOUSE_EVENT;
		send_mouse(w, mouse_x, mouse_y, mouse_b);
	}
}

static void
handle_message(unsigned long win, char *data)
{
	union dmessage *message = (union dmessage *)data;
	union dcommand command;
	struct window *window = find_window(win);
	
	if (!window) {
		return;
	}
	switch (message->type) {
	case DMESSAGE_REQ_BOUNDS:
		window->bounds = message->req_bounds.bounds;
		window->origin = message->req_bounds.origin;
		window->bounds.x1 += window->origin.x;
		window->bounds.y1 += window->origin.y;
		window->bounds.x2 += window->origin.x;
		window->bounds.y2 += window->origin.y;
		command.type = DCOMMAND_SET_BOUNDS;
		command.set_bounds.win = win;
		command.set_bounds.bounds = message->req_bounds.bounds;
		command.set_bounds.origin = message->req_bounds.origin;
		write(dinx_fd, &command, sizeof(union dcommand));
		break;		
	case DMESSAGE_REQ_MAP:
		command.type = DCOMMAND_MAP;
		command.window.win = win;
		write(dinx_fd, &command, sizeof(union dcommand));
		break;
	case DMESSAGE_REQ_UNMAP:
		command.type = DCOMMAND_UNMAP;
		command.window.win = win;
		write(dinx_fd, &command, sizeof(union dcommand));
		break;
	case DMESSAGE_REQ_RAISE:
		if (!window->nextabove) {
			break;
		}
		/* remove from the list */
		window->nextabove->nextbelow = window->nextbelow;
		if (window->nextbelow) {
			window->nextbelow->nextabove = window->nextabove;
		} else {
			bottomchild = window->nextabove;
			bottomchild->nextbelow = NULL;
		}
		/* add to the top of the list */
		window->nextabove = NULL;
		window->nextbelow = topchild;
		topchild->nextabove = window;
		topchild = window;

		command.type = DCOMMAND_RAISE;
		command.window.win = win;
		write(dinx_fd, &command, sizeof(union dcommand));
		break;
	case DMESSAGE_REQ_LOWER:
		if (!window->nextbelow) {
			break;
		}
		/* remove from the list */
		window->nextbelow->nextabove = window->nextabove;
		if (window->nextabove) {
			window->nextabove->nextbelow = window->nextbelow;
		} else {
			topchild = window->nextbelow;
			topchild->nextabove = NULL;
		}
		/* add to the bottom of the list */
		window->nextbelow = NULL;
		window->nextabove = bottomchild;
		bottomchild->nextbelow = window;
		bottomchild = window;

		command.type = DCOMMAND_LOWER;
		command.window.win = win;
		write(dinx_fd, &command, sizeof(union dcommand));
		break;
	default:
		fprintf(stderr, "unexpected message type %d\n", message->type);
		break;
	}
}

void handle_key(int key)
{
	struct window *w;
	
	for (w = topchild; w; w = w->nextbelow) {
		union dmessage message;
	
		message.type = DMESSAGE_KEY;
		message.key.key = key;
		dinx_message_to(dinx_fd, &message, w->win);
	}
}

void handle_mouse(int x, int y, int b)
{
	struct window *w;
	
	for (w = topchild; w; w = w->nextbelow) {
		if (event_count(w->win) == 0) {
			send_mouse(w, x, y, b);
		} else {
			w->flags |= NEEDS_MOUSE_EVENT;
		}
	}
}

void open_mouse(void)
{
	mouse_fd = open(mouse_filename, O_RDONLY | O_NONBLOCK);
	if (mouse_fd < 0) {
		perror(mouse_filename);
	}
}

void close_mouse(void)
{
	if (mouse_fd >= 0) {
		close(mouse_fd);
	}
	mouse_fd = -1;
}

void update_mouse(void)
{
	static signed char data[5];
	static int count = 0;
	int i, changed = 0;

	while (1) {
		i = read(mouse_fd, data + count, 5 - count);
		if (i <= 0) {
			break;
		}
		count += i;
		if (count == 5) {
			count = 0;
			changed = 1;
			mouse_b = (~data[0]) & 0x07;
			mouse_x += (int)data[1] + (int)data[3];
			mouse_y -= (int)data[2] + (int)data[4];
		}
	}
	if (changed) {
		mouse_x = max(mouse_x, 0);
		mouse_x = min(mouse_x, fb_var.xres-1);
		mouse_y = max(mouse_y, 0);
		mouse_y = min(mouse_y, fb_var.yres-1);
		handle_mouse(mouse_x, mouse_y, mouse_b);
	}
}

void release(int sig)
{
	dinx_simple_command(dinx_fd, DCOMMAND_DISABLE);
	close_mouse();
	vt_ack_release();
}

void acquire(int sig)
{
	vt_ack_acquire();
	vt_wait_acquire();
	dinx_simple_command(dinx_fd, DCOMMAND_ENABLE);
	dinx_simple_command(dinx_fd, DCOMMAND_REFRESH);
	open_mouse();
}

void quit(int sig)
{
	fprintf(stderr, "dinxd: killed by signal %d\n", sig);
	exit(sig);
}

int main(int argc, char **argv)
{
	int c;
	struct stat st;
	
	while (1) {
		c = getopt_long_only(argc, argv, "dfhv", options, NULL);
		if (c == -1) {
			break;
		}
		switch (c) {
		case 0:
			break;
		case 'd':
			svr_filename = optarg;
			break;
		case 'f':
			framebuffer = optarg;
			break;
		case 'v':
			printf("dinx-" VERSION "\n");
			exit(0);
		case 'h':
		default:
			printf(usage_message);
			exit(0);
		}
	}
	signal(SIGHUP, quit);
	signal(SIGINT, quit);
	signal(SIGQUIT, quit);
	signal(SIGTERM, quit);

	if (stat(svr_filename, &st)) {
		perror(svr_filename);
		exit(1);
	}
	if (!S_ISCHR(st.st_mode)) {
		fprintf(stderr, "%s: not a character special device\n",
				svr_filename);
		exit(1);
	}
	if (major(st.st_rdev) != DINX_MAJOR) {
		fprintf(stderr, "%s: major number does not match dinx.h\n",
				svr_filename);
		exit(1);
	}
	if (!(minor(st.st_rdev) & DINX_MINOR_SVR_BIT)) {
		fprintf(stderr, "%s: minor number is not a server node\n",
				svr_filename);
		exit(1);
	}
	dinx_fd = open(svr_filename, O_RDWR);
	if (dinx_fd < 0) {
		perror(svr_filename);
		vt_close();
		exit(1);
	}
	if (vt_init(release, acquire)) {
		exit(1);
	}
	atexit(vt_close);
	if (fb_init(framebuffer)) {
		exit(1);
	}
	if (0 > dinx_select_fb(dinx_fd, fb_number)) {
		if (errno == EINVAL) {
			fprintf(stderr, "invalid framebuffer %d\n", fb_number);
		} else if (errno == ENOSYS) {
			fprintf(stderr, "no blitter for framebuffer\n");
		} else {
			perror("dinx_select_fb");
		}
		exit(1);
	}
	dinx_simple_command(dinx_fd, DCOMMAND_ENABLE);
	dinx_simple_command(dinx_fd, DCOMMAND_REFRESH);
	mouse_x = fb_var.xres / 2;
	mouse_y = fb_var.yres / 2;
	mouse_b = 0;
	open_mouse();
	atexit(close_mouse);

	while (1) {
		struct pollfd pfd[2];
		
		pfd[0].fd = dinx_fd;
		pfd[0].events = POLLIN;
		pfd[1].fd = mouse_fd;
		pfd[1].events = POLLIN;
		pfd[2].fd = vt_fd;
		pfd[2].events = POLLIN;
		poll(pfd, 3, -1);

		for (; pfd[0].revents & POLLIN; poll(&pfd[0], 1, 0)) {
			union devent event;
			union dcommand command;
			struct window *window;
			
			event.type = DEVENT_NONE;
			read(dinx_fd, &event, sizeof(union devent));
			switch (event.type) {
			case DEVENT_MESSAGE:
				handle_message(event.message.win, event.message.data);
				break;
			case DEVENT_WIN_OPENED:
				window = malloc(sizeof(struct window));
				window->win = event.win_status.win;
				window->bounds.x1 = 0;
				window->bounds.y1 = 0;
				window->bounds.x2 = 0;
				window->bounds.y2 = 0;
				window->origin.x = 0;
				window->origin.y = 0;
				window->flags = NEEDS_MOUSE_EVENT;
				/* new windows go on top */
				window->nextabove = NULL;
				window->nextbelow = topchild;
				if (topchild) {
					topchild->nextabove = window;
				} else {
					bottomchild = window;
				}
				topchild = window;
				send_tree_info(window->win);
				feed_window(window->win);
				/* Currently we use this default to suit PC hardware */
				command.type = DCOMMAND_PREFER_REDRAW;
				command.window.win = window->win;
				write(dinx_fd, &command, sizeof(union dcommand));
				break;
			case DEVENT_WIN_CLOSED:
				window = find_window(event.win_status.win);
				if (!window) {
					break;
				}
				if (window->nextabove) {
					window->nextabove->nextbelow = window->nextbelow;
				} else {
					topchild = window->nextbelow;
					if (topchild) {
						topchild->nextabove = NULL;
					}
				}
				if (window->nextbelow) {
					window->nextbelow->nextabove = window->nextabove;
				} else {
					bottomchild = window->nextabove;
					if (bottomchild) {
						bottomchild->nextbelow = NULL;
					}
				}
				free(window);
				break;
			case DEVENT_WIN_EMPTY:
				feed_window(event.win_status.win);
				break;
			default:
				fprintf(stderr, "unexpected event type %d\n", event.type);
				break;
			}
		}
		if ((pfd[1].revents & POLLIN) && (mouse_fd >= 0)) {
			update_mouse();
		}
		for (; pfd[2].revents & POLLIN; poll(&pfd[2], 1, 0)) {
			char c;
			
			if (1 == read(vt_fd, &c, 1)) {
				handle_key(c);
			}
		}
	}
	return 0;
}



